#include                <string.h>
#include                "tap.h"
#include                "libFireBird.h"

#define PROGRAM_NAME    "StandbyMode"
#define VERSION         "V0.5"
#define ROOTDIR          "/ProgramFiles/Settings/"PROGRAM_NAME
#define INIFILE         PROGRAM_NAME ".ini"
#define LOGFILE         PROGRAM_NAME ".log"
#define TIMERNAME       PROGRAM_NAME ".rec"

TAP_ID                  (0x8E0A424a);
TAP_PROGRAM_NAME        (PROGRAM_NAME" "VERSION);
TAP_PROGRAM_VERSION     (VERSION);
TAP_AUTHOR_NAME         ("FireBird");
TAP_DESCRIPTION         ("StandbyMode");
TAP_ETCINFO             (__DATE__);

typedef enum
{
  ST_Init,
  ST_WaitCheckTimeout,
  ST_Idle
} tState;

tState                  State = ST_Init;
char                    Log[255];
dword                   ActiveStandbyStart;
dword                   ActiveStandbyEnd;
int                     GracePeriod;
bool                    DoShutdown;
dword                   WaitTimeout;

void CreateSettingsDir(void)
{
  HDD_TAP_PushDir();
  HDD_ChangeDir("/ProgramFiles");
  if(!TAP_Hdd_Exist("Settings")) TAP_Hdd_Create("Settings", ATTR_FOLDER);
  HDD_ChangeDir("Settings");
  if(!TAP_Hdd_Exist("StandbyMode")) TAP_Hdd_Create("StandbyMode", ATTR_FOLDER);
  HDD_TAP_PopDir();
}

void WriteLog(char *s)
{
  LogEntryGeneric(PROGRAM_NAME, TRUE, s);
}

void LoadINI(void)
{
  HDD_TAP_PushDir();
  TAP_Hdd_ChangeDir(ROOTDIR);
  if(INIOpenFile(INIFILE, PROGRAM_NAME) == INILOCATION_NewFile)
  {
    ActiveStandbyStart = 0x0800;
    INISetString("Start", "0800");

    ActiveStandbyEnd = 0x173b;
    INISetString("End", "2359");

    GracePeriod = 60;
    INISetInt("GracePeriod", 60);

    INISaveFile(INIFILE, INILOCATION_AtCurrentDir, PROGRAM_NAME);
  }
  else
  {
    ActiveStandbyStart = INIGetInt("Start", 800, 0, 2400);
    ActiveStandbyStart = ((int)(ActiveStandbyStart / 100) << 8) | (ActiveStandbyStart % 100);

    ActiveStandbyEnd = INIGetInt("End", 2359, 0, 2400);
    ActiveStandbyEnd = ((int)(ActiveStandbyEnd / 100) << 8) | (ActiveStandbyEnd % 100);

    GracePeriod = INIGetInt("GracePeriod", 60, 0, 600);
  }
  INICloseFile();
  HDD_TAP_PopDir();
}

void ForeignTimer(bool isForeignTimer)
{
  HDD_TAP_PushDir();
  TAP_Hdd_ChangeDir(ROOTDIR);
  INIOpenFile(INIFILE, PROGRAM_NAME);
  INISetInt("ForeignTimer", isForeignTimer ? 1 : 0);
  INISaveFile(INIFILE, INILOCATION_AtCurrentDir, PROGRAM_NAME);
  INICloseFile();
  HDD_TAP_PopDir();
}

bool isForeignTimer(void)
{
  bool ret;

  HDD_TAP_PushDir();
  TAP_Hdd_ChangeDir(ROOTDIR);
  INIOpenFile(INIFILE, PROGRAM_NAME);
  ret = INIGetInt("ForeignTimer", 0, 0, 1) != 0;
  INICloseFile();
  HDD_TAP_PopDir();

  return ret;
}

bool SetStandbyMode(bool Active, bool VFD, bool Tuner, bool SCART)
{
  byte                 *__etcInfo;
  byte                  ModeWord;
  static void (*__Appl_WriteEeprom)(bool) = NULL;

  __etcInfo = (byte*)FIS_vEtcInfo();
  if(!__etcInfo) return FALSE;

  if(!__Appl_WriteEeprom)
  {
    __Appl_WriteEeprom = (void*)TryResolve("_Z16Appl_WriteEepromb");
    if(!__Appl_WriteEeprom) return FALSE;
  }

  if(Active)
    ModeWord = 0x88 | (VFD ? 0x01 : 0x00) | (Tuner ? 0x02 : 0x00) | (SCART ? 0x04 : 0x00);
  else
    ModeWord = 0x00;

  if(ModeWord != __etcInfo[0x2b])
  {
    __etcInfo[0x2b] = ModeWord;
    __Appl_WriteEeprom(FALSE);
  }

  return TRUE;
}

bool GetStandbyMode(bool *Active)
{
  byte                 *__etcInfo;

  __etcInfo = (byte*)FIS_vEtcInfo();
  if(!__etcInfo) return FALSE;

  *Active = (__etcInfo[0x2b] != 0);

  return TRUE;
}

bool GetActiveModeAccordingToTime(void)
{
  word                  mjd;
  byte                  Hour, Min, Sec;
  dword                 Time;
  bool                  Active;

  TAP_GetTime(&mjd, &Hour, &Min, &Sec);
  Time = (Hour << 8) | Min;

  Active = ((Time >= ActiveStandbyStart) && (Time < ActiveStandbyEnd));

  return Active;
}

void CheckStandbyMode(void)
{
  word                  mjd;
  byte                  Hour, Min, Sec;
  static byte           LastMin = 0;
  bool                  CurrentActiveMode, NewActiveMode;

  //Only check once a minute
  TAP_GetTime(&mjd, &Hour, &Min, &Sec);
  if(Min != LastMin)
  {
    NewActiveMode = GetActiveModeAccordingToTime();
    if(GetStandbyMode(&CurrentActiveMode) && (CurrentActiveMode != NewActiveMode))
    {
      if(NewActiveMode)
      {
        SetStandbyMode(TRUE, TRUE, FALSE, FALSE);
        WriteLog("Switched into active standby mode");
      }
      else
      {
        SetStandbyMode(FALSE, FALSE, FALSE, FALSE);
        WriteLog("Switched into passive standby mode");
      }
      if(DoShutdown) Shutdown(TaskPower);
    }
    LastMin = Min;
  }
}

void RemoveStandbyModeTimer(void)
{
  TYPE_TimerInfo        TimerInfo;
  int                   i, NrTimer;
  word                  Year;
  byte                  Month, Day, WeekDay;

  NrTimer = TAP_Timer_GetTotalNum();
  for(i = NrTimer - 1; i >= 0; i--)
  {
    TAP_Timer_GetInfo(i, &TimerInfo);
    if(!strcmp(TimerInfo.fileName, TIMERNAME))
    {
      TAP_ExtractMjd(TimerInfo.startTime >> 16, &Year, &Month, &Day, &WeekDay);
      TAP_SPrint(Log, "Removed a StandbyMode timer for %4.4d-%2.2d-%2.2d %2.2d:%2.2d", Year, Month, Day, (TimerInfo.startTime >> 8) & 0xff, TimerInfo.startTime & 0xff);
      WriteLog(Log);
      TAP_Timer_Delete(i);
    }
  }
}

void CreateStandbyModeTimer(void)
{
  TYPE_TimerInfo        TimerInfo;
  word                  mjd;
  byte                  Hour, Min, Sec;
  dword                 Time, SwitchTime, TimerEndTime;
  int                   i, NrTimer;
  word                  Year;
  byte                  Month, Day, WeekDay;
  tAutoDescrambleTimer  ADTimer;

  ForeignTimer(TRUE);

  //Calculate next switch Time
  TAP_GetTime(&mjd, &Hour, &Min, &Sec);
  Time = (Hour << 8) | Min;

  if(Time < ActiveStandbyStart)
    SwitchTime = (mjd << 16) | ActiveStandbyStart;
  else if(Time < ActiveStandbyEnd)
    SwitchTime = (mjd << 16) | ActiveStandbyEnd;
  else
    SwitchTime = ((mjd + 1) << 16) | ActiveStandbyStart;

  //Check if there is an older timer which expands into the switch time or
  //  if there is an newer timer which is within one hour of the switch time
  NrTimer = TAP_Timer_GetTotalNum();
  for(i = 0; i < NrTimer; i++)
  {
    TAP_Timer_GetInfo(i, &TimerInfo);
    TimerEndTime = AddTime(TimerInfo.startTime, TimerInfo.duration);
    if(((TimerInfo.startTime <= SwitchTime) && (TimerEndTime >= SwitchTime)) ||
       ((TimerInfo.startTime > SwitchTime) && (TimeDiff(SwitchTime, TimerInfo.startTime) < GracePeriod)))
    {
      TAP_ExtractMjd(TimerInfo.startTime >> 16, &Year, &Month, &Day, &WeekDay);
      TAP_SPrint(Log, "Timer '%s' @ %4.4d-%2.2d-%2.2d %2.2d:%2.2d used for switching", TimerInfo.fileName, Year, Month, Day, (TimerInfo.startTime >> 8) & 0xff, TimerInfo.startTime & 0xff);
      WriteLog(Log);
      return;
    }
  }

  //Same check with the AD timer
  if(FlashADGetInfo(&ADTimer))
  {
    if(((ADTimer.StartTime <= SwitchTime) && (ADTimer.EndTime >= SwitchTime)) ||
       ((ADTimer.StartTime > SwitchTime) && (TimeDiff(SwitchTime, ADTimer.StartTime) < GracePeriod)))
    {
      TAP_ExtractMjd(ADTimer.StartTime >> 16, &Year, &Month, &Day, &WeekDay);
      TAP_SPrint(Log, "AD timer @ %4.4d-%2.2d-%2.2d %2.2d:%2.2d used for switching", Year, Month, Day, (ADTimer.StartTime >> 8) & 0xff, ADTimer.StartTime & 0xff);
      WriteLog(Log);
      return;
    }
  }

  //If not, create a new P-timer at X, lasting for 2 minutes
  TimerInfo.isRec           = 0;
  TimerInfo.tuner           = 0;
  TimerInfo.svcType         = SVC_TYPE_Tv;
  TimerInfo.svcNum          = 0;
  TimerInfo.reservationType = RESERVE_TYPE_Onetime;
  TimerInfo.nameFix         = 1;
  TimerInfo.duration        = 2;
  TimerInfo.startTime       = SwitchTime;
  strcpy(TimerInfo.fileName, TIMERNAME);

  int ret = TAP_Timer_Add(&TimerInfo);

  TAP_ExtractMjd(TimerInfo.startTime >> 16, &Year, &Month, &Day, &WeekDay);
  if(ret)
  {
    TAP_SPrint(Log, "Failed to created a StandbyMode timer @ %4.4d-%2.2d-%2.2d %2.2d:%2.2d. Error code = 0x%x", Year, Month, Day, (TimerInfo.startTime >> 8) & 0xff, TimerInfo.startTime & 0xff, ret);
    WriteLog(Log);
  }
  else
  {
    TAP_SPrint(Log, "Created a StandbyMode timer @ %4.4d-%2.2d-%2.2d %2.2d:%2.2d", Year, Month, Day, (TimerInfo.startTime >> 8) & 0xff, TimerInfo.startTime & 0xff);
    WriteLog(Log);
  }
  ForeignTimer(FALSE);
}

void CheckRecording(void)
{
  static word           LastMin = 99;
  TYPE_RecInfo          RecInfo;
  word                  mjd;
  byte                  hour, min, sec;
  dword                 i;

  //Check only if DoShutdown is active
  if(!DoShutdown) return;

  //Check only once a minutes
  TAP_GetTime(&mjd, &hour, &min, &sec);
  if(min == LastMin) return;

  LastMin = min;

  for(i = 0; i < HDD_NumberOfRECSlots(); i++)
  {
    TAP_Hdd_GetRecInfo(i, &RecInfo);
    if(RecInfo.recType == RECTYPE_Normal)
    {
      DoShutdown = FALSE;
      break;
    }
  }
}

dword TAP_EventHandler(word event, dword param1, dword param2)
{
  (void) param2;

  if(event == EVT_IDLE)
  {
    switch(State)
    {
      case ST_Init:
      {
        RemoveStandbyModeTimer();
        State = ST_WaitCheckTimeout;
        WaitTimeout = TAP_GetTick() + 1500;
        break;
      }

      case ST_WaitCheckTimeout:
      {
        if(TAP_GetTick() > WaitTimeout) State = ST_Idle;
        break;
      }

      case ST_Idle:
      {
        CheckStandbyMode();
        CheckRecording();
        break;
      }
    }
  }

  if(event == EVT_KEY)
  {
    if(DoShutdown) WriteLog("No shutdown due to key event");
    DoShutdown = FALSE;
  }

  if(event == EVT_STOP && param1 < 2)
  {
    CreateStandbyModeTimer();
    TAP_Exit();
  }

  return param1;
}

int TAP_Main(void)
{
  bool ForeignTimer;

  CreateSettingsDir();
  LoadINI();

  ForeignTimer = isForeignTimer();
  if(ForeignTimer)
    WriteLog("No shutdown as this is a foreign timer");
  else if(BootReason() != BOOT_TIMER)
  {
    TAP_SPrint(Log, "No shutdown because wakeup is ");
    switch(BootReason())
    {
      case BOOT_FRONT_PANEL:    strcat(Log, "front panel"); break;
      case BOOT_REMOTE_CONTROL: strcat(Log, "remote control"); break;
      case BOOT_TIMER:          strcat(Log, "timer"); break;
      case BOOT_CRASH_REBOOT:   strcat(Log, "crash reboot"); break;
      default:                  strcat(Log, "unknown"); break;
    }
    WriteLog(Log);
  }

  DoShutdown = (BootReason() == BOOT_TIMER) && !isForeignTimer();

  return 1;
}
